import sys
import shutil
import tempfile
import subprocess as sp
import string
import requests

import httpx
import trio

from datetime import datetime
from pathlib import Path
from typing import List, Dict, Optional
from functools import partial, wraps
from time import time

import click
from colorama import init as colorama_init
from colorama import Fore, Back, Style
from halo import Halo

from ..config.github import config_github, config_announce_grade

from ..utils.github_scanner import github_headers, get_github_endpoint_paged_list_async
from ..utils.google_student import Gstudents


# We will use {source_repo}/{source_tmpl_issue} as the template to give students feedback

# The Pull request we made would fetch the first issue which has the same title
# as patch_branch to be the content

def measure_time(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        return result, te-ts
    return wrap


@click.command()
@click.argument('homework_prefix')
@click.option('--token', default=config_github['personal_access_token'], help="github access token")
@click.option('--org', default=config_github['organization'], show_default=True)
@click.option('--only-id', nargs=1, help="only id to announce")
@click.option('--feedback-source-repo', default=config_announce_grade['feedback_source_repo'], show_default=True)
def announce_grade(homework_prefix, token, org, only_id, feedback_source_repo):
    '''announce student grades to each hw repo'''

    # TODO: use logging lib to log messages
    colorama_init()
    spinner = Halo(stream=sys.stderr)

    student_feedback_title = f"Grade for {homework_prefix}"

    gstudents = Gstudents()
    feedback_vars = gstudents.left_join(homework_prefix)

    # Clone feedback repo & set needed variables
    cur = Path('.')

    for d in cur.glob("feedback-tmp-*"):
        shutil.rmtree(d)
    spinner.info("delete dated folder")

    root_folder = Path(tempfile.mkdtemp(
        prefix="feedback-tmp-{}-".format(datetime.now().strftime("%b%d%H%M%S")), dir="."))
    spinner.succeed(
        f"Create tmp folder {Fore.YELLOW}{root_folder}{Style.RESET_ALL}")

    feedback_repo_path = root_folder / 'feedbacks'

    spinner.start(f"cloning feeback source repo : {feedback_source_repo}")
    _, t = measure_time(sp.run)(['git', 'clone',
                                 f'https://github.com/{org}/{feedback_source_repo}.git', feedback_repo_path.name, ], cwd=root_folder, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
    spinner.succeed(
        f"cloning feeback source repo : {feedback_source_repo} ... {t:4.2f} sec")
    client = httpx.AsyncClient(headers=httpx.Headers({
        "User-Agent": "GitHubClassroomUtils/1.0",
        "Authorization": "token " + token,
        # needed for the check-suites request
        "Accept": "application/vnd.github.antiope-preview+json"
    }))

    hw_path = feedback_repo_path / homework_prefix / 'reports'

    # generate feedbacks
    fbs, t = measure_time(gen_feedbacks)(
        homework_prefix, hw_path, feedback_vars)
    spinner.succeed(f"Generate content for feedbacks ... {t:5.3f} sec")

    # handle only_id
    if only_id:
        try:
            # detect possible buggy condition
            info = gstudents.get_student(only_id)
        except RuntimeError as e:
            print(' *='*30)
            print('Warning!')
            print(e)
            return
        only_repo_name = get_hw_repo_name(
            homework_prefix, info['github_handle'])
        fbs = list(filter(lambda fb: fb['repo_name'] == only_repo_name, fbs))

    async def push_to_remote(feedback_title, feedbacks):
        # push to remote
        async def push_feedback(fb):
            request_body = {
                'title': feedback_title,
                'body': fb['value']
            }
            try:
                issue_num = await find_existing_issue(client, org, fb['repo_name'], feedback_title)
            except BaseException as e:
                print(f'error on {fb["repo_name"]}')
                return
            if issue_num:
                request_body['state'] = 'open'  # reopen issue
                url = f"https://api.github.com/repos/{org}/{fb['repo_name']}/issues/{issue_num}"
                await edit_issue_async(client, url, issue_num, request_body)
            else:
                url = f"https://api.github.com/repos/{org}/{fb['repo_name']}/issues"
                await create_issue_async(client, url, request_body)
            print(f'success {fb["repo_name"]}')
        async with trio.open_nursery() as nursery:
            for fb in feedbacks:
                nursery.start_soon(push_feedback, fb)

    _, t = measure_time(trio.run)(push_to_remote, student_feedback_title, fbs)
    spinner.succeed(f"Push feedbacks to remote ... {t:5.2f} sec")
    spinner.succeed(f'finished announce grade')
    return


def get_hw_repo_name(hw_prefix, gh_handle):
    return f'{hw_prefix}-{gh_handle}'


def gen_feedbacks(hw_prefix, fb_root_path, fb_vars) -> List[Dict]:
    # substitude with template info
    result = []
    for vars in fb_vars:
        feedback = {'repo_name': None, 'value': None}
        # generate names for student h.w. repos
        student_id, handle = vars['student_id'], vars['github_handle']
        repo_name = get_hw_repo_name(hw_prefix, handle)
        feedback['repo_name'] = repo_name
        # fetch each students feedback template file by id
        fb_tmpl_file = fb_root_path/f'{student_id}.md'
        if fb_tmpl_file.exists():
            # substitude value inside feedback template
            with open(fb_tmpl_file, 'r') as tmpl_file:
                fb_tmpl = string.Template(tmpl_file.read())
                feedback['value'] = fb_tmpl.safe_substitute(vars)

        if feedback['repo_name'] and feedback['value']:
            result.append(feedback)
    return result


async def find_existing_issue(client, org, repo_name, issue_title) -> Optional[int]:
    # TODO: remove argument: org
    # print(f'finding existing issue... :{repo_name}/{issue_title}')
    issues = await get_github_endpoint_paged_list_async(
        client, f"repos/{org}/{repo_name}/issues", state='all')
    # TODO: deal with repo not exist case
    if len(issues) != 0:
        try:
            issue = next(i for i in issues if i['title'] == issue_title)
            return int(issue['number'])
        except:
            pass
    return None


async def edit_issue_async(client, url, issue_number, requst_body):
    res = await client.patch(url, json=requst_body)
    if res.status_code != 200:
        print(f'failed on editing issue: {url}')
        try:
            print(f"{res.json()['errors'][0]['message']}")
        except:
            pass


async def create_issue_async(client, url, request_body):
    res = await client.post(url, json=request_body)
    if res.status_code != 201:
        print(f'failed on {url}')
        try:
            print(f"{res.json()['errors'][0]['message']}")
        except:
            pass

if __name__ == "__main__":
    announce_grade()



import click
import trio
from tqdm import tqdm

from typing import List

from colorama import init as colorama_init
from colorama import Fore, Back, Style
from ..utils.github_scanner import query_matching_repos
from ..utils.github_entities import Team

from ..config.github import config_github, config_grant_read_access

# Use a "team slug"
# for example: "2019 Teaching-team" -> "2019_teaching-team"
@click.command()
@click.argument('hw_title')
@click.option('--token', default=config_github['personal_access_token'], help="github access token")
@click.option('--org', default=config_github['organization'], show_default=True)
@click.option('--team', default=config_grant_read_access['reader_team_slug'], show_default=True)
def grant_read_access(hw_title, token, org, team):
    '''Grant read access right of TA's group to students' homework repo'''
    print("start script")
    colorama_init()
    teaching_team = Team(org, team, token)

    repos = query_matching_repos(org,
                                 github_repo_prefix=f'{hw_title}-',
                                 github_token=token,
                                 verbose=True)
    repo_names = [ r['name'] for r in repos ]

    # show repos to operate on
    ncols = 3
    cols = [ repo_names[i::ncols] for i in range(ncols)]
    print('repos to operate on:')
    for a,b,c in zip(*cols):
        print(f'  {a:<30}{b:<30}{c:<}')

    builder = pbar_builder()
    builder.set_config(total=len(repo_names))
    with builder.build(desc='fired') as fired, builder.build(desc='returned') as returned:
        async def subscribe_to_repo(team:Team, repo_name:str):
            res = await team.add_team_repository_async(repo_name, permission="pull")
            if res.status_code != 204:
                print(res)
            returned.update(1)

        async def async_github():
            async with trio.open_nursery() as nursery:
                for repo_name in repo_names:
                    nursery.start_soon(subscribe_to_repo, teaching_team, repo_name)
                    fired.update(1)
                    fired.refresh()
                return
        trio.run(async_github)

class pbar_builder():
    def __cover_dict(self, src, dst):
        for k,v in src.items():
            dst[k] = v
    def __init__(self):
        self.config = {}
        self.next_po = 0

    def set_config(self,**kwargs):
        self.__cover_dict(kwargs, self.config)

    def build(self, **kwargs):
        cfg = {}
        self.__cover_dict(self.config, cfg)
        self.__cover_dict(kwargs, cfg )
        cfg['position'] = self.next_po
        self.next_po += 1
        return tqdm(**cfg)

if __name__ == "__main__":
    grant_read_access()



# Initialize google spreadsheet constants
from pathlib import Path
from configparser import ConfigParser

from .config_util import init_config_file

config_gsheet = None

def init_constants():
    '''initialize all constants'''
    global config_gsheet
    template_name = 'google_sheet.ini'
    cfile = "gsheet_config.ini"

    if not Path(cfile).exists():
        init_config_file(template_name, cfile)

    config = ConfigParser()
    config.read(cfile)

    # export global variables
    config_gsheet = dict(config['google_spreadsheet'])



import click
import sys
from typing import List, Dict

import requests
from requests.models import Response

from colorama import init as colorama_init
from colorama import Fore, Back, Style
from halo import Halo

from ..config.github import config_github, config_add_students
from ..utils.github_scanner import get_github_endpoint, github_headers
from ..utils.github_entities import Team


def print_table(data, cols=5, wide=15, indent=2):
    '''Prints formatted data on columns of given width.'''
    n, r = divmod(len(data), cols)
    pat = '{{:{}}}'.format(wide)
    line = '\n{}'.format(' '*indent).join(pat * cols for _ in range(n))

    # indent the first line
    line = ' '*indent + line

    last_line = ' '*indent + pat * r
    print(line.format(*data))
    print(last_line.format(*data[n*cols:]))


# Use a "team slug"
# for example: "2019 Students-hello" -> "2019_students-hello"

@click.command()
@click.argument('student_handles', nargs=-1)
@click.option('--token', default=config_github['personal_access_token'], help="github access token")
@click.option('--org', default=config_github['organization'], show_default=True)
@click.option('--team', default=config_add_students['default_team_slug'], show_default=True)
def add_students(student_handles, token, org, team):
    '''
    student_handles: github user to add (usernames)
    '''
    if len(student_handles) == 0:
        print(f'required handles')
        return 1
    colorama_init()

    github_students = student_handles
    github_organization = org
    github_team = team
    github_token = token

    if github_token == "":
        print(Back.RED + Fore.BLACK +
              "No github_token is given, use "
              + Fore.WHITE + "cmd_argument" + Fore.BLACK
              + " or sepcify it in "
              + Fore.WHITE + "github_congif.py "
              + Back.RESET + Fore.RESET)
    #print('org: {}'.format(github_organization))
    #print('token: {}'.format(github_token))
    # print("students:{}".format(github_students))

    with Halo() as spinner:
        spinner.text = "fetch existing team members from GitHub"
        # spinner.start()

        team = Team(github_organization, team_slug=github_team,
                    github_token=github_token)
        spinner.succeed()
        dic = {
            'team': github_team,
            'n_mem': len(team.members.keys()),
            'num_style': Fore.GREEN,
            'reset_str': Back.RESET + Fore.RESET
        }
        spinner.info(
            "{team} - {num_style}{n_mem}{reset_str} members".format(**dic))
        spinner.text_color = "green"

    existed_members = set(team.members.keys())
    outside_users = list(set(github_students)-existed_members)

    #print("Users to invite:")
    #print_table(outside_users, cols=5, wide=15)

    invalid_id = []
    with Halo() as spinner:
        spinner.text = ""
        # spinner.start()
        total = len(outside_users)
        for idx, u in enumerate(outside_users, start=1):
            spinner.text = "{}/{} Check valid GitHub username : {}".format(
                idx, total, u)
            if check_is_github_user(u, github_token) == False:
                invalid_id.append(u)
        spinner.text = "{}/{} Check valid GitHub username".format(total, total)
        spinner.succeed()

    if len(invalid_id) != 0:
        print("Find non-existed github user names:")
        # control strings take space
        print_table([Fore.RED+i+Fore.RESET for i in invalid_id],
                    cols=5, wide=25)

    non_member_valid_users = list(set(outside_users) - set(invalid_id))

    # membership info
    membership_infos = {key: "unknown" for key in non_member_valid_users}
    with Halo() as spinner:
        spinner.text = "Check Membership information"
        spinner.start()
        total = len(non_member_valid_users)
        for idx, username in enumerate(non_member_valid_users, start=1):
            spinner.text = "{}/{} Check Membership information : {}".format(
                idx, total, username)
            res = team.get_memberships(username)
            if res.status_code == 200:
                membership_infos[username] = res.json()['state']
        spinner.text = "{}/{} Check Membership information".format(
            total, total)
        spinner.succeed()

    pending_users = [u for u in membership_infos.keys(
    ) if membership_infos[u] == "pending"]
    no_memship_users = [
        u for u in membership_infos.keys() if membership_infos[u] == "unknown"]

    print("Users already in pending state (total:{}):".format(len(pending_users)))
    print_table(pending_users)

    print("Users to add (total: {})".format(len(no_memship_users)))
    print_table(no_memship_users)

    failed_users = []
    with Halo() as spinner:
        for user_name in no_memship_users:
            spinner.text = Fore.GREEN + "adding user: " + \
                Fore.RESET+"{}".format(user_name)
            spinner.start()
            res = team.add_user_to_team(user_name)
            if res.status_code == 200:
                spinner.succeed()
            else:
                failed_users.append(user_name)
                spinner.text += ", return code: " + Fore.RED + \
                    str(res.status_code) + Fore.RESET
                spinner.fail()
    failed_users = list(set(failed_users))

    if len(failed_users) != 0:
        print("Users failed to add")
        print_table(failed_users)

    with Halo() as spinner:
        spinner.text = "Adding students successfully"
        spinner.info()


def check_is_github_user(github_id, github_token) -> bool:
    res = requests.get("https://api.github.com/users/{}".format(github_id),
                       headers=github_headers(github_token))
    if res.status_code == 200:
        return True
    else:
        return False


if __name__ == "__main__":
    add_students()



# import click

from pathlib import Path
from configparser import ConfigParser

from .config_util import init_config_file

# These variables will be set in init_constants()
config_github = None
config_add_students = None
config_event_times = None
config_announce_grade = None
config_crawl_classroom = None
config_grant_read_access = None

# will be call inside invisible_hand/__init__.py


def init_constants():
    '''initialize all constants'''
    global config_github
    global config_announce_grade
    global config_add_students
    global config_event_times
    global config_crawl_classroom
    global config_grant_read_access

    template_name = 'github.ini'
    cfile = "github_config.ini"

    if not Path(cfile).exists():
        init_config_file(template_name, cfile)
        # print('propagate config files')
        # sys.exit(1)
    config = ConfigParser()
    config.read(cfile)

    # export global variables
    config_github = dict(config['github'])
    config_event_times = dict(config['event_times'])
    config_add_students = dict(config['add_students'])
    config_announce_grade = dict(config['announce_grade'])
    config_crawl_classroom = dict(config['crawl_classroom'])
    config_grant_read_access = dict(config['grant_read_access'])



from .utils.google_student import Gstudents
import click

from prompt_toolkit import prompt

from .scripts.add_students import add_students
from .scripts.event_times import event_times
from .scripts.patch_project import patch_project
from .scripts.announce_grade import announce_grade
from .scripts.crawl_classroom import crawl_classroom
from .scripts.grant_read_access import grant_read_access


@click.group(invoke_without_command=True, context_settings=dict(help_option_names=["-h", "--help"]))
@click.pass_context
def cli(ctx):
    '''Toolkits for github courses
    '''
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


cli.add_command(add_students)
cli.add_command(event_times)
cli.add_command(patch_project)
cli.add_command(announce_grade)
cli.add_command(crawl_classroom)
cli.add_command(grant_read_access)


def test():

    # g = Gsheet(spreadsheetId='1WlbKDBdhC03m_mFikWwwveDvSQKwKKo20SO9KlbeSqQ')
    # g.playground()

    g = Gstudents()

    # print( g.gsheet.sheets().get(sid) )
    # print(gsheet.get_values())
    # print(gsheet.get_values(A1=SAMPLE_RANGE_NAME))
    # print(f'title list: {gsheet.get_title_list()}')

def main():
    cli()

if __name__ == "__main__":
    # test()
    cli()



import sys
import requests
import asyncio
import trio
import httpx

from typing import List

from requests.models import Response

from colorama import init as colorama_init
from colorama import Fore, Back, Style
from halo import Halo

from ..config.github import config_github
from .github_scanner import get_github_endpoint, github_headers, get_github_endpoint_paged_list

def main():
    print ("nothin to run")

class Team:
    def __init__(self, org, team_slug="", github_token=""):
        self.info_sucess = False
        self.id = ""
        self.org = org
        self.team_slug = team_slug
        self.members = dict()
        self.github_token = github_token
        self._get_members()
        colorama_init()


        # async
        self.httpx_headers = httpx.Headers({
            "User-Agent": "GitHubClassroomUtils/1.0",
            "Authorization": "token " + github_token,
            # needed for the check-suites request
            "Accept": "application/vnd.github.antiope-preview+json"
        })
        self.async_client = httpx.AsyncClient(headers=self.httpx_headers)

    def add_user_to_team(self, user_name) -> Response:
        res = requests.put(
            "https://api.github.com/teams/{}/memberships/{}".format(
                self.id, user_name),
            headers=github_headers(self.github_token)
        )
        return res

    def add_team_repository(self, repo, permission="pull") -> Response:
        '''If permission=pull, it is equivalent to make this team subscribe to this repo 
        '''
        api_param = {
            'team_id': self.id,
            'owner': self.org,
            'repo': repo
        }
        return requests.put("https://api.github.com/teams/{team_id}/repos/{owner}/{repo}".format(**api_param),
                            headers=github_headers(self.github_token))

    async def add_team_repository_async(self, repo, permission="pull") -> Response:
        '''If permission=pull, it is equivalent to make this team subscribe to this repo 
        '''
        api_param = {
            'team_id': self.id,
            'owner': self.org,
            'repo': repo
        }
        # print(f'param: {api_param}')
        return await self.async_client.put("https://api.github.com/teams/{team_id}/repos/{owner}/{repo}".format(**api_param))

    def get_memberships(self, user_name: str) -> Response:
        return requests.get("https://api.github.com/teams/{}/memberships/{}".format(self.id, user_name),
                            headers=github_headers(self.github_token))

    def _require_team_id(func):
        def decorator(self):
            if self.id == "":
                self._get_team_id()
            func(self)
        return decorator

    @_require_team_id
    def _get_members(self):
        try:
            # res = get_github_endpoint(
            #     endpoint="teams/{}/members".format(self.id),
            #     github_token=self.github_token,
            # )
            res = get_github_endpoint_paged_list(
                endpoint="teams/{}/members".format(self.id),
                github_token=self.github_token,
                verbose=False
            )
        except:
            print(Back.RED + Fore.BLACK +
                  "Cannot fetch team id"
                  + Back.RESET + Fore.RESET)
            print("Team info:")
            print("    org: {}".format(self.org))
            print("    team_slug: {}".format(self.team_slug))
            sys.exit(1)
        for user in res:
            self.members[user['login']] = user
        # print(res)

    def _get_team_id(self):
        try:
            result = get_github_endpoint(
                "orgs/{}/teams/{}".format(self.org, self.team_slug),
                config_github['personal_access_token'],
            )
        except:
            print(Back.RED + Fore.BLACK +
                  "Cannot fetch team id"
                  + Back.RESET + Fore.RESET)
            print("Team info:")
            print("    org: {}".format(self.org))
            print("    team_slug: {}".format(self.team_slug))
            sys.exit(1)
        self.id = result['id']



if __name__ == "__main__":
    main()


# github_event_times.py
# Dan Wallach <dwallach@rice.edu>
# Available subject to the Apache 2.0 License
# https://www.apache.org/licenses/LICENSE-2.0

#import argparse
import click
from pprint import pprint
from datetime import datetime, timedelta
import re
import sys
from pathlib import Path
from typing import List, NamedTuple, Tuple

from halo import Halo
from colorama import init as colorama_init
from colorama import Fore, Back, Style
from tabulate import tabulate

from ..config.github import config_github, config_event_times

from ..utils.github_scanner import LOCAL_TIMEZONE
from ..utils.github_scanner import *
from ..utils.github_entities import Team


def is_deadline_passed(dl: datetime, submit: datetime) -> Tuple[bool, timedelta]:
    dl = dl.astimezone(submit.tzinfo)
    if dl < submit:
        return True, submit - dl
    else:
        return False, dl-submit


class commitInfo(NamedTuple):
    commit_hash: str  # 7 characters hash value
    pushed_time: str
    msg: str
    repo: str


class GitHubRepoInfo(NamedTuple):
    name: str
    commit_hash: str  # 7 characters hash value

# python3 github_event_times.py hw0-ianre657:cb75e99
@click.command()
@click.argument('input-file')
@click.option('--token', default=config_github['personal_access_token'], help="github access token")
@click.option('--org', default=config_github['organization'], show_default=True)
@click.option('--deadline', default=config_event_times['deadline'], show_default=True,
              help='deadline format: "yyyy-mm-dd" or any ISO8601 format string (timezone will be set to local timezone)')
@click.option('--target-team', default=None, help="specific team to operate on")
def event_times(input_file, org, token, deadline, target_team):
    '''
    input-file: file contains list of repo-hash.

    repo-hash : string in <repo>:<hash> format
            hw0-ianre657:cb75e99
    '''
    global github_organization
    global github_token

    try:
        parsed_repos = get_repo_infos(input_file)
    except FileNotFoundError as e:
        print(str(e))
        return

    colorama_init(autoreset=True)
    spinner = Halo(stream=sys.stderr)

    github_organization = org
    github_token = token

    print(f'deadline: {deadline}')

    submit_deadline = iso8601.parse_date(deadline)
    submit_deadline = submit_deadline.replace(tzinfo=LOCAL_TIMEZONE)

    spinner.info(f"Deadline : {submit_deadline}")
    success_group = []
    fail_group = []
    spinner.start("Start to check late submissions")

    # get team membershup info
    if target_team is not None:
        only_team_members = set(Team(org=github_organization, team_slug=target_team, github_token=github_token).members.keys())

    for idx, repo in enumerate(parsed_repos, start=1):
        #print("get commit time for {}".format(repo))
        if target_team is not None:
            import re
            user_id = re.sub('hw[\d]+-', '', repo.name )
            # print(f'user_id :{user_id}')
            if user_id not in only_team_members:
                continue
        spinner.text = f"({idx}/{len(parsed_repos)}) Checking {repo.name}"
        result = getRepoCommitTime(
            org=github_organization, repo=repo.name, commit_hash=repo.commit_hash)
        for r in result:
            # print(r)
            passed, delta = is_deadline_passed(
                submit_deadline, iso8601.parse_date(r.pushed_time))
            if passed:
                fail_group.append({
                    'repo-name': r.repo,
                    'commit-hash': r.commit_hash,
                    'time-passed': delta,
                    'last-pushtime': r.pushed_time
                })
            else:
                success_group.append((r, delta))
                #print(f'{r}: {delta} later')
    spinner.succeed("Check finished")
    print('='*20, 'REPORT', '='*20)
    print(f'Total submissions : {len(parsed_repos)}')
    print(f'late submissions: {len(fail_group)}')
    print(f'Submission Deadline: {submit_deadline}')
    print(tabulate(fail_group, headers="keys"))

def get_repo_infos( filename: str ) -> List[GitHubRepoInfo]:
    st = Path(filename).read_text()
    infos = st.split()
    result = []
    for r in infos:
        repo, commit_hash = r.split(":")
        result.append(GitHubRepoInfo(repo, commit_hash))
    return result


def getRepoCommitTime(org: str, repo: str, commit_hash: str) -> List[commitInfo]:
    global github_token
    response = get_github_endpoint_paged_list(f"repos/{org}/{repo}/events", github_token, verbose=False)
    event_list = [x for x in response if x['type'] == 'PushEvent']
    # find the localtiome of the given commit SHA that is pushed.
    msgs = []
    for event in event_list:
        try:
            github_id = event['actor']['login']
            date = localtime_from_iso_datestr(event['created_at'])
            target_commits = [x for x in event['payload']
                              ['commits'] if x['sha'][0:7] == commit_hash]

            if len(target_commits) == 0:
                continue
            elif len(target_commits) >= 2:
                print("SHA Conflict for user: {}".format(github_id))
            else:
                target_commit = target_commits[0]

                commit_message = tex_escape(target_commit['message'].splitlines()[
                                            0])  # only the first line if multiline
                commit_hash = target_commit['sha'][0:7]

                result = commitInfo(commit_hash, date, commit_message, repo)
                # result_str = "{} sha:{}, msg:{}, date:{}".format(github_id, commit_hash, commit_message, date)
                msgs.append(result)
        except KeyError:
            print("Error: malformed event!")
            pprint(event)
    return msgs

# https://stackoverflow.com/questions/16259923/how-can-i-escape-latex-special-characters-inside-django-templates


def tex_escape(text: str) -> str:
    """
        :param text: a plain text message
        :return: the message escaped to appear correctly in LaTeX
    """
    conv = {
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\^{}',
        '\\': r'\textbackslash{}',
        '<': r'\textless{}',
        '>': r'\textgreater{}',
    }
    regex = re.compile('|'.join(re.escape(str(key))
                                for key in sorted(conv.keys(), key=lambda item: - len(item))))
    return regex.sub(lambda match: conv[match.group()], text)


if __name__ == "__main__":
    event_times()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invisible_hand',
 'invisible_hand.config',
 'invisible_hand.config.templates',
 'invisible_hand.scripts',
 'invisible_hand.utils']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.3,<0.5.0',
 'gitpython>=3.1.7,<4.0.0',
 'google-api-python-client>=1.10.0,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'halo>=0.0.30,<0.0.31',
 'httpx>=0.14.1,<0.15.0',
 'ipython>=7.17.0,<8.0.0',
 'iso8601>=0.1.12,<0.2.0',
 'lxml>=4.5.2,<5.0.0',
 'pandas>=1.1.1,<2.0.0',
 'prompt-toolkit>=3.0.6,<4.0.0',
 'pygsheets>=2.0.3,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'selenium>=3.141.0,<4.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tqdm>=4.48.2,<5.0.0',
 'trio>=0.16.0,<0.17.0',
 'xlsxwriter>=1.3.3,<2.0.0']

entry_points = \
{'console_scripts': ['invisible-hand = invisible_hand.__main__:main']}

setup_kwargs = {
    'name': 'invisible-hand',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Ian Chen',
    'author_email': 'ianre657@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, List, Optional

from xlsxwriter.utility import xl_col_to_name
import pygsheets

from ..config.gsheet import config_gsheet

class pygsheetInteractor:
    def __init__(self,pyg=pygsheets):
        self.gc = pyg.authorize()
        self.sht = None

    def open_by_url(self, url):
        self.sht = self.gc.open_by_url(url)

    def _get_wks_by_title(self, title):
        if self.sht is None:
            raise RuntimeError('Call open_by_url before using any function')
        return self.sht.worksheet_by_title(title)

    def get_all_record(self, title, head=1) -> List[Dict]:
        wks = self._get_wks_by_title(title)
        endtag = f'{xl_col_to_name(wks.cols-1)}{wks.rows}'

        grange = pygsheets.GridRange(end=endtag)

        data = wks.get_values(
            grange=grange,
            include_tailing_empty=True,
            include_tailing_empty_rows=False
        )

        idx = head-1
        keys = data[idx]
        field_index_map = {field: n for n,
                        field in enumerate(keys) if field.strip() != ""}
        vals = []
        for row in data[idx+1:]:
            val = {field: row[i] for field, i in field_index_map.items()}
            vals.append(val)
        return vals

class Gstudents:
    def __init__(self, url=config_gsheet['spreadsheet_url'], actor=None):
        self.actor = actor if actor is not None else pygsheetInteractor()
        self.actor.open_by_url(url)

        self.config = {
            'main_wks_title': 'StudentInfo',
            'main_wks_required_fields':
                [
                    'github_handle',
                    'student_id',
                    'name',
                    'email',
                ],
            'main_kws_key_field': 'student_id'
        }

    def get_students(self) -> List[Dict[str, str]]:
        return self.actor.get_all_record(self.config['main_wks_title'])

    def get_student(self, student_id ) -> Optional[Dict]:
        '''Fetch a single students info, will traverse the entire table
            Use get_students manually if you need to get each student's info
        '''
        students = self.get_students()
        ret = []
        for s in students:
            if student_id == s['student_id']:
                ret.append(s)

        if len(ret) > 1:
            raise RuntimeError(f'duplicate student_id :{ret}')
        return ret[0] if len(ret) > 0 else None

    def left_join(self, right_sheet_title) -> List[Dict[str, str]]:
        # TODO: check id collision
        left_dicts = self.get_students()
        right_dicts = self.actor.get_all_record(right_sheet_title)

        def left_matching(left, right, key_field: str):
            for l in left:
                for r in right:
                    if l[key_field] == r[key_field]:
                        yield l, r
        joined_infos = []
        for left, right in left_matching(left_dicts, right_dicts, self.config['main_kws_key_field']):
            student = {**left, **right}
            joined_infos.append(student)
        return joined_infos
